# Author: Dylan Turner
# Description: Handle OpenCV processing of images

from cv2 import \
    VideoCapture, resize, namedWindow, moveWindow, imshow, waitKey, \
    destroyAllWindows, imread, BORDER_CONSTANT, copyMakeBorder, resize, imread, \
    GaussianBlur
from numpy import shape, uint8
from cvzone import stackImages
from cvzone.SelfiSegmentationModule import SelfiSegmentation
from time import sleep

WIN_TITLE = 'Feed'

class Cam:
    def __init__(self, settings):
        # Set up WebCam access and output
        self._settings = settings

        self._vid_feed = VideoCapture(settings.camera)
        self._vid_feed.set(3, settings.screen_width)
        self._vid_feed.set(4, settings.screen_height)

        _success, baseFrame = self._vid_feed.read()
        if not _success:
            print('Failed to read from camera!')
            quit()
        _height, _width, self.channels = baseFrame.shape

        if not settings.disable_win:
            namedWindow(WIN_TITLE)

        # Set up removal
        self._segmentor = SelfiSegmentation()

        if settings.bg_img != '':
            self._bg_img = self._create_correctly_sized_bg()
        else:
            self._bg_img = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._vid_feed.release()
        destroyAllWindows()

    # Return regular camera frame and with the bg removal applied, as well as them put together
    def frames(self):
        success, frame = self._vid_feed.read()

        if not success:
            return (None, None, None)

        if shape(self._bg_img) == ():
            no_bg_frame = self._segmentor.removeBG(
                frame, self._settings.fill_color,
                threshold = self._settings.rm_thresh
            )
        else:
            no_bg_frame = self._segmentor.removeBG(
                frame, self._bg_img,
                threshold = self._settings.rm_thresh
            )

        if self._settings.blur:
            removed_bg = frame - no_bg_frame
            blurred = GaussianBlur(removed_bg, (77, 77), 21)
            no_bg_frame += blurred

        return (frame, no_bg_frame, self._stack_frames(frame, no_bg_frame))

    def display(self, frame):
        if not self._settings.disable_win:
            imshow(WIN_TITLE, frame)

        if waitKey(1) & 0xFF == ord('q'):
            return False

        return True

    def _stack_frames(self, left_frame, right_frame):
        new_width = int(self._settings.screen_width * self._settings.view_scale)
        new_height = int(self._settings.screen_height * self._settings.view_scale)
        return stackImages(
            [
                resize(left_frame, (new_width, new_height)),
                resize(right_frame, (new_width, new_height))
            ], 1, 1
        )

    # Adjust self._bg_img to be the size we want
    def _create_correctly_sized_bg(self):
        bg_img = imread(self._settings.bg_img)

        # Scale to match y and be centered
        bg_height, bg_width, _channels = bg_img.shape
        aspect = float(bg_width) / float(bg_height)
        new_width = int(self._settings.screen_height * aspect)
        bg_img = resize(bg_img, (new_width, self._settings.screen_height))

        # Scale down
        if new_width > self._settings.screen_width:
            startx = int((new_width - self._settings.screen_width) / 2)
            endx = new_width - startx
            bg_img = bg_img[0:self._settings.screen_height, startx:endx]
        else:
            padding = int((self._settings.screen_width - new_width) / 2)
            bg_img = copyMakeBorder(bg_img, 0, 0, padding, padding, BORDER_CONSTANT)

        # Make sure correct size
        bg_img = resize(bg_img, (self._settings.screen_width, self._settings.screen_height))

        return bg_img




# Author: Dylan Turner
# Description: File-to-file version of Cam class

from cv2 import VideoCapture, VideoWriter, VideoWriter_fourcc, namedWindow, destroyAllWindows
from cvzone.SelfiSegmentationModule import SelfiSegmentation
from bgrm.cam import Cam

WIN_TITLE = 'Conversion'

class VidFile(Cam):
    def __init__(self, settings):
        self._settings = settings

        # Set up file input
        self._vid_feed = VideoCapture(settings.input_file)
        _success, base_frame = self._vid_feed.read()
        height, width, self.channels = base_frame.shape
        self._settings.screen_width = width
        self._settings.screen_height = height
        settings.screen_width = width
        settings.screen_height = height

        if not settings.disable_win:
            namedWindow(WIN_TITLE)

        # Set up removal
        self._segmentor = SelfiSegmentation()

        if settings.bg_img != '':
            self._bg_img = self._create_correctly_sized_bg()
        else:
            self._bg_img = None

        # Set up file output
        self._out_feed = VideoWriter(
            settings.output_file, VideoWriter_fourcc('M', 'J', 'P', 'G'),
            settings.fps, (width, height)
        )

    def __exit__(self, exc_type, exc_value, traceback):
        self._vid_feed.release()
        self._out_feed.release()
        destroyAllWindows() 

    def write(self, frame):
        self._out_feed.write(frame)




#!/usr/bin/env python3

# Author: Dylan Turner
# Description: Main loop for bgrm app

from numpy import shape
from bgrm.cam import Cam
from bgrm.virt_cam import VirtCam
from bgrm.vid_file import VidFile
from bgrm.settings import AppSettings

def main():
    settings = AppSettings.from_cli()

    if not settings.file_mode:
        virt_dev_name = '/dev/video' + str(settings.virt_dev)
        with Cam(settings) as cam, VirtCam(virt_dev_name) as virt_cam:
            virt_cam.format(cam, settings)
            while True:
                _frame, no_bg_frame, stacked_frames = cam.frames()
                virt_cam.write(no_bg_frame)

                if not cam.display(stacked_frames):
                    break
    else:
        with VidFile(settings) as vid_file:
            while True:
                frame, no_bg_frame, stacked_frames = vid_file.frames()

                if shape(frame) == ():
                    break

                vid_file.write(no_bg_frame)
                
                if not vid_file.display(stacked_frames):
                    break




# Author: Dylan Turner
# Description: Keep app settings in one place

from dataclasses import dataclass
from argparse import ArgumentParser
from subprocess import run, PIPE

@dataclass
class AppSettings:
    virt_dev: int
    camera: int
    screen_width: int
    screen_height: int
    view_scale: float
    rm_thresh: float
    fill_color: tuple[float, float, float]
    bg_img: str
    blur: bool
    disable_win: bool
    file_mode: bool
    input_file: str
    output_file: str
    fps: int

    @staticmethod
    def from_cli():
        virt_dev = pull_virt_dev()
        args = cli_args()

        # Convert fill color from string to tuple
        fill_col = args.color.split(',')
        if len(fill_col) != 3:
            print('Failed to parse fill color argument \'' + args.color + '\'')
            raise SettingsError
        fill_col = tuple([ float(e) for e in fill_col ])

        return AppSettings(
            virt_dev, args.camera, args.width, args.height, args.scale,
            args.thresh, fill_col, args.bg, args.blur, args.disable_window,
            args.file_mode, args.input, args.output, args.fps
        )

# Get video settings from command line
def pull_virt_dev():
    shell_result = run(
        [ 'ls', '/sys/devices/virtual/video4linux' ],
        stderr = PIPE, stdout = PIPE, text = True
    )
    raw_devs = shell_result.stdout.split('\n')
    raw_devs = list(filter(lambda elem: elem != '', raw_devs))
    
    devs = [ int(vid_str.replace('video', '')) for vid_str in raw_devs ]

    if len(devs) < 1:
        print('Failed to find virtual device! Do you have v4l2loopback set up?')
        raise SettingsError

    return devs[0]

def cli_args():
    parser = ArgumentParser()

    # Set up arguments
    parser.add_argument(
        '-c', '--camera', type = int, default = 0,
        help = 'ID of camera device to use for input'
    )
    parser.add_argument(
        '-b', '--bg', type = str, default = '',
        help = 'Background image'
    )
    parser.add_argument(
        '-T', '--thresh', type = float, default = 0.4,
        help = 'Background removal threshold'
    )
    parser.add_argument(
        '-w', '--width', type = int, default = 1280,
        help = 'Width of camera input'
    )
    parser.add_argument(
        '-H', '--height', type = int, default = 960,
        help = 'Height of camera input'
    )

    # Extra options
    parser.add_argument(
        '--blur', help = 'Blur background (overrides --bg)',
        action = 'store_true'
    )
    parser.add_argument(
        '--disable_window', help = 'Disable feedback window',
        action = 'store_true'
    )
    parser.add_argument(
        '-C', '--color', type = str, default = '0,0,0',
        help = 'List of R, G, B to replace background with',
    )

    # File versions
    parser.add_argument(
        '--file_mode',
        help = 'Remove the background from video files instead of cameras (overrides --camera)',
        action = 'store_true'
    )
    parser.add_argument(
        '-i', '--input', type = str, default = '',
        help = 'Only used with file mode. Input file to remove background',
    )
    parser.add_argument(
        '-o', '--output', type = str, default = '',
        help = 'Only used with file mode. Output file to store removed background feed.'
    )
    parser.add_argument(
        '-f', '--fps', type = int, default = 30,
        help = 'Only used with file mode. Output file fps'
    )

    # Mostly useless window options
    parser.add_argument(
        '-s', '--scale', type = float, default = 0.5,
        help = 'Scale factor from camera size to window.'
    )

    return parser.parse_args()




# Author: Dylan Turner
# Description: Encapsulate code to format and write to virtual camera device

from cv2 import cvtColor, COLOR_BGR2YUV_I420
from v4l2 import \
        v4l2_format, V4L2_BUF_TYPE_VIDEO_OUTPUT, V4L2_FIELD_NONE, \
    V4L2_PIX_FMT_YUV420, VIDIOC_S_FMT
from fcntl import ioctl

class VirtCam:
    def __init__(self, dev_name):
        self._device = open(dev_name, 'wb')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._device.close()

    def write(self, frame):
        self._device.write(cvtColor(frame, COLOR_BGR2YUV_I420))

    def format(self, actual_cam, settings):
        # Setup format info for writing to cam device
        fmt = v4l2_format()
        fmt.type = V4L2_BUF_TYPE_VIDEO_OUTPUT
        fmt.fmt.pix.field = V4L2_FIELD_NONE
        fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420
        fmt.fmt.pix.width = settings.screen_width
        fmt.fmt.pix.height = settings.screen_height
        fmt.fmt.pix.bytesperline = settings.screen_width * actual_cam.channels
        fmt.fmt.pix.sizeimage = \
            settings.screen_width * settings.screen_height * actual_cam.channels

        # Set device format
        print('Format loopback format result (0 good): {}'.format(
            ioctl(self._device, VIDIOC_S_FMT, fmt)
        ))




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bgrm']

package_data = \
{'': ['*']}

install_requires = \
['cvzone==1.5.3',
 'mediapipe==0.8.11',
 'opencv-python==4.6.0.66',
 'protobuf==3.19.0',
 'v4l2-python3==0.3.1']

setup_kwargs = {
    'name': 'bgrm',
    'version': '9',
    'description': 'Remove backgrounds from video feeds in your web cam applications.',
    'long_description': '# Background Remover\n\n## The Need\n\nIt\'s been good long while since Microsoft first released a Teams version for Linux and yet, one of Teams\' coolest features doesn\'t exist in said Linux version: removable backgrounds. As someone who uses Linux for their daily driver, this annoys me.\n\nWell, I\'m an engineer, so of course, I found a solution.\n\nUsing OpenCV and a v4l2loopback device (basically a virtual webcam you can write data to), I threw together a Python application that takes your normal webcam input, removes and replaces the background, and outputs that to the created video device. Problem solved :)\n\nNote, this will work anywhere WebCams are used, not just Teams\n\nNow, the program can also be used to remove backgrounds from video files and save them as video files as well!\n\n## How to Use\n\n### WebCam Replacement\n\nDependencies:\n    - python >= 3.8 (3.10 is what\'s supported officially)\n    - pip\n    - v4l2loopback\n\nSetup:\n1. Configure v4l2loopback (may not be necessary):\n    - Recommended something like this:\n    ```\n    export DEVICE_ARR=(`ls /sys/devices/virtual/video4linux | tr -d \'video\'`); \\\n    sudo modprobe v4l2loopback \\\n        devices=1 exclusive_caps=1 video_nr=${DEVICE_ARR[1]} max_buffers=2 \\\n        card_label=v4l2lo\n    ```\n2. Install with `pip install bgrm`\n\nThen, you can run: \n- Run with `python -m bgrm <options>` (use `--help` to see all options)\n- Example: `python -m bgrm -b ~/Pictures/Wallpapers/ni-skyline-wallpaper.png -w 320 -H 240 -s 2.0`\n\n### File Replacement\n\nYou can also remove the background from video files. It works just like the WebCam, but instead of setting the `--camera` cli arg, you call the program like this:\n\n`python -m bgrm --file_mode -i <input file> -o <output file> <other options>`\n\n## Build from Repo\n\nYou can also build the package yourself from source (or grab the latest version from the releases tab)\n\nTo do that you need the "poetry" build system.\n\nRun `poetry build` and install the whl from the dist/ folder\n',
    'author': 'Dylan Turner',
    'author_email': 'dylantdmt@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/blueOkiris/bgrm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
